#include "nBodySim.hpp"
#include "tree.hpp"
#include <omp.h>
#include <cmath>
#include <string>
#include <fstream>
#include <iomanip>

nBodySim::nBodySim(std::string datafile) {
    std::string dummy;
    std::ifstream file(datafile);
    file >> dummy >> dummy >> nParticles_;

    masses_ = new double[nParticles_];
    for (unsigned i=0; i < nParticles_; ++i) {
        file >> masses_[i];
    }

    double maxCoord = 0;
    positions_ = new double[3*nParticles_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < nParticles_; ++j) {
            file >> positions_[i+3*j];
            if (std::abs(positions_[i+3*j]) > maxCoord) maxCoord = std::abs(positions_[i+3*j]);
        }
    }
    
    velocities_ = new double[3*nParticles_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < nParticles_; ++j) {
            file >> velocities_[i+3*j];
        }
    }

    for (unsigned i=0; i < nParticles_; ++i) {
        file >> softening_;
    }
    
    potential_ = new double[nParticles_];
    for (unsigned i=0; i < nParticles_; ++i) {
        file >> potential_[i];
    }

    forces_ = new double[3*nParticles_];
    #pragma omp parallel for
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        forces_[i] = 0.0;
    }
    
    file.close();

    tree_ = new Tree(positions_, forces_, nParticles_, 4*maxCoord, new double[3]{0, 0, 0});
}

nBodySim::~nBodySim() {
    delete[] masses_;
    delete[] positions_;
    delete[] velocities_;
    delete[] potential_;
    delete[] forces_;
}

void nBodySim::runSimulation(double dt, unsigned nSteps) {
    for (unsigned i=0; i < nSteps; ++i) {
        doTimeStep(dt);
    }
}

void nBodySim::calculateForces() {
    #pragma omp parallel for
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        forces_[i] = 0.0;
    }
    // loop over each pair of particles and calculate the force between them
    #pragma omp parallel for // schedule(guided)
    for (unsigned i = 0; i < nParticles_; i++) {
        double mi = masses_[i];
        double s = meanInterparticleDistance_; // softening_[i];
        double px = positions_[3*i];
        double py = positions_[3*i+1];
        double pz = positions_[3*i+2];
        double fx = 0.0;
        double fy = 0.0;
        double fz = 0.0;
        for (unsigned j = i+1; j < nParticles_; j++) {
            // if (j <= i) continue;
            // calculate distance between particles
            double dx = positions_[3*j]   - px;
            double dy = positions_[3*j+1] - py;
            double dz = positions_[3*j+2] - pz;
            double r2 = dx*dx + dy*dy + dz*dz;
            double r = std::sqrt(r2);
            dx /= r;
            dy /= r;
            dz /= r;
            // calculate forces
            double mj = masses_[j];
            double fxij = mi*mj * dx / (r2 + s*s);
            double fyij = mi*mj * dy / (r2 + s*s);
            double fzij = mi*mj * dz / (r2 + s*s);
            fx += fxij;
            fy += fyij;
            fz += fzij;
            forces_[3*j]   -= fxij;
            forces_[3*j+1] -= fyij;
            forces_[3*j+2] -= fzij;
        }
        forces_[3*i]   += fx;
        forces_[3*i+1] += fy;
        forces_[3*i+2] += fz;
    }
}

void nBodySim::treeCalculateForces() {
    // TODO
}

double nBodySim::calculateMeanInterparticleDistance() {
    meanInterparticleDistance_ = 0.0;
    // loop over each pair of particles and calculate the distance between them
    #pragma omp parallel for reduction(+:meanInterparticleDistance_) // schedule(guided, 64)
    for (unsigned i = 0; i < nParticles_; i++) {
        double px = positions_[3*i];
        double py = positions_[3*i+1];
        double pz = positions_[3*i+2];
        double localMeanInterparticleDistance = 0.0;
        for (unsigned j = i+1; j < nParticles_; j++) {
            // calculate distance between particles
            double dx = positions_[3*j]   - px;
            double dy = positions_[3*j+1] - py;
            double dz = positions_[3*j+2] - pz;
            double r2 = dx*dx + dy*dy + dz*dz;
            double r = std::sqrt(r2);
            localMeanInterparticleDistance += r;
        }
        meanInterparticleDistance_ += localMeanInterparticleDistance;
    }
    meanInterparticleDistance_ /= (nParticles_*(nParticles_-1)/2);
    return meanInterparticleDistance_;
}

double nBodySim::calculateMeanForceMagnitude() {
    double forceMagnitude = 0.0;
    #pragma omp parallel for reduction(+:forceMagnitude)
    for (unsigned i = 0; i < nParticles_; i++) {
        double fx = forces_[3*i];
        double fy = forces_[3*i+1];
        double fz = forces_[3*i+2];
        double f2 = fx*fx + fy*fy + fz*fz;
        double f = std::sqrt(f2);
        forceMagnitude += f;
    }
    meanForceMagnitude_ = forceMagnitude / nParticles_;
    return meanForceMagnitude_;
}

void nBodySim::doTimeStep(double dt) {
    #pragma omp parallel for
    for (unsigned i=0; i < nParticles_; ++i) {
        for (unsigned j=0; j < 3; ++j) {
            // TODO
            tree_->drift(dt/2);
            tree_->update();
            tree_->kick(dt);
            tree_->drift(dt/2);
        }
    }
}

void nBodySim::write2file(const double* array, std::string filename, unsigned dim) const {
    // writes array to file, with N rows and dim columns
    std::ofstream file(filename);
    for (unsigned i = 0; i < nParticles_; i++) {
        for (unsigned j = 0; j < dim; j++) {
            file << std::right << std::setw(16) << array[dim*i+j] << " ";
        }
        file << std::endl;
    }
}

void nBodySim::saveState2file(unsigned step, std::string filename) const {
    std::ofstream file(filename);
    // save step number
    file << "step = " << step;
    // save number of particles
    file << nParticles_ << std::endl;
    // save masses
    for (unsigned i=0; i < nParticles_; ++i) {
        file << masses_[i] << std::endl;
    }
    // save positions
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        file << positions_[i] << std::endl;
    }
    // save velocities
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        file << velocities_[i] << std::endl;
    }
    // save softening
    for (unsigned i=0; i < nParticles_; ++i) {
        file << softening_ << std::endl;
    }
    // save potential
    for (unsigned i=0; i < nParticles_; ++i) {
        file << potential_[i] << std::endl;
    }
    file.close();
}